@@ -3,14 +3,14 @@
 //! These are mainly created via the methods provided by `EncodeExt` or `DecodeExt` traits.
 use std;
 use std::cmp;
+use std::fmt;
+use std::iter;
 use std::marker::PhantomData;
 use std::mem;
 
-pub use chain::{Buffered, DecoderChain, EncoderChain};
-
 use bytes::BytesEncoder;
 use marker::Never;
-use {ByteCount, Decode, Encode, EncodeExt, Eos, Error, ErrorKind, ExactBytesEncode, Result};
+use {ByteCount, Decode, Encode, EncodeExt, Eos, Error, ErrorKind, Result, SizedEncode};
 
 /// Combinator for converting decoded items to other values.
 ///
@@ -22,6 +22,21 @@ pub struct Map<D, T, F> {
     _item: PhantomData<T>,
 }
 impl<D: Decode, T, F> Map<D, T, F> {
+    /// Returns a reference to the inner decoder.
+    pub fn inner_ref(&self) -> &D {
+        &self.inner
+    }
+
+    /// Returns a mutable reference to the inner decoder.
+    pub fn inner_mut(&mut self) -> &mut D {
+        &mut self.inner
+    }
+
+    /// Takes ownership of this instance and returns the inner decoder.
+    pub fn into_inner(self) -> D {
+        self.inner
+    }
+
     pub(crate) fn new(inner: D, map: F) -> Self
     where
         F: Fn(D::Item) -> T,
@@ -40,25 +55,49 @@ where
 {
     type Item = T;
 
-    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<(usize, Option<Self::Item>)> {
-        track!(self.inner.decode(buf, eos)).map(|(n, r)| (n, r.map(&self.map)))
+    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<usize> {
+        track!(self.inner.decode(buf, eos))
+    }
+
+    fn finish_decoding(&mut self) -> Result<Self::Item> {
+        let item = track!(self.inner.finish_decoding())?;
+        Ok((self.map)(item))
     }
 
     fn requiring_bytes(&self) -> ByteCount {
         self.inner.requiring_bytes()
     }
+
+    fn is_idle(&self) -> bool {
+        self.inner.is_idle()
+    }
 }
 
 /// Combinator for modifying encoding/decoding errors.
 ///
 /// This is created by calling `{DecodeExt, EncodeExt}::map_err` method.
 #[derive(Debug)]
-pub struct MapErr<C, F, E> {
+pub struct MapErr<C, E, F> {
     inner: C,
     map_err: F,
     _error: PhantomData<E>,
 }
-impl<C, F, E> MapErr<C, F, E> {
+impl<C, E, F> MapErr<C, E, F> {
+    /// Returns a reference to the inner encoder or decoder.
+    pub fn inner_ref(&self) -> &C {
+        &self.inner
+    }
+
+    /// Returns a mutable reference to the inner encoder or decoder.
+    pub fn inner_mut(&mut self) -> &mut C {
+        &mut self.inner
+    }
+
+    /// Takes ownership of this instance and returns the inner encoder or decoder.
+    pub fn into_inner(self) -> C {
+        self.inner
+    }
+
     pub(crate) fn new(inner: C, map_err: F) -> Self
     where
         F: Fn(Error) -> E,
@@ -71,7 +110,7 @@ impl<C, F, E> MapErr<C, F, E> {
         }
     }
 }
-impl<D, F, E> Decode for MapErr<D, F, E>
+impl<D, E, F> Decode for MapErr<D, E, F>
 where
     D: Decode,
     F: Fn(Error) -> E,
@@ -79,17 +118,27 @@ where
 {
     type Item = D::Item;
 
-    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<(usize, Option<Self::Item>)> {
+    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<usize> {
         self.inner
             .decode(buf, eos)
             .map_err(|e| (self.map_err)(e).into())
     }
 
+    fn finish_decoding(&mut self) -> Result<Self::Item> {
+        self.inner
+            .finish_decoding()
+            .map_err(|e| (self.map_err)(e).into())
+    }
+
     fn requiring_bytes(&self) -> ByteCount {
         self.inner.requiring_bytes()
     }
+
+    fn is_idle(&self) -> bool {
+        self.inner.is_idle()
+    }
 }
-impl<C, F, E> Encode for MapErr<C, F, E>
+impl<C, E, F> Encode for MapErr<C, E, F>
 where
     C: Encode,
     F: Fn(Error) -> E,
@@ -117,9 +166,9 @@ where
         self.inner.is_idle()
     }
 }
-impl<C, F, E> ExactBytesEncode for MapErr<C, F, E>
+impl<C, E, F> SizedEncode for MapErr<C, E, F>
 where
-    C: ExactBytesEncode,
+    C: SizedEncode,
     F: Fn(Error) -> E,
     Error: From<E>,
 {
@@ -160,28 +209,22 @@ where
 {
     type Item = D1::Item;
 
-    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<(usize, Option<Self::Item>)> {
-        let offset = if self.inner1.is_none() {
-            let (size, item) = track!(self.inner0.decode(buf, eos))?;
-            if let Some(d) = item.map(&self.and_then) {
-                self.inner1 = Some(d);
-            }
-            size
-        } else {
-            0
-        };
-        if let Some(result) = self.inner1
-            .as_mut()
-            .map(|d| track!(d.decode(&buf[offset..], eos)))
-        {
-            let (size, item) = result?;
-            if item.is_some() {
-                self.inner1 = None;
-            }
-            Ok((offset + size, item))
-        } else {
-            Ok((offset, None))
+    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<usize> {
+        let mut offset = 0;
+        if self.inner1.is_none() {
+            bytecodec_try_decode!(self.inner0, offset, buf, eos);
+            let item = track!(self.inner0.finish_decoding())?;
+            self.inner1 = Some((self.and_then)(item));
         }
+
+        let inner1 = self.inner1.as_mut().expect("Never fails");
+        bytecodec_try_decode!(inner1, offset, buf, eos);
+        Ok(offset)
+    }
+
+    fn finish_decoding(&mut self) -> Result<Self::Item> {
+        let mut d = track_assert_some!(self.inner1.take(), ErrorKind::IncompleteDecoding);
+        track!(d.finish_decoding())
     }
 
     fn requiring_bytes(&self) -> ByteCount {
@@ -191,6 +234,10 @@ where
             self.inner0.requiring_bytes()
         }
     }
+
+    fn is_idle(&self) -> bool {
+        self.inner1.as_ref().map_or(false, |d| d.is_idle())
+    }
 }
 
 /// Combinator for converting items into ones that
@@ -204,6 +251,21 @@ pub struct MapFrom<E, T, F> {
     from: F,
 }
 impl<E, T, F> MapFrom<E, T, F> {
+    /// Returns a reference to the inner encoder.
+    pub fn inner_ref(&self) -> &E {
+        &self.inner
+    }
+
+    /// Returns a mutable reference to the inner encoder.
+    pub fn inner_mut(&mut self) -> &mut E {
+        &mut self.inner
+    }
+
+    /// Takes ownership of this instance and returns the inner encoder.
+    pub fn into_inner(self) -> E {
+        self.inner
+    }
+
     pub(crate) fn new(inner: E, from: F) -> Self {
         MapFrom {
             inner,
@@ -235,9 +297,9 @@ where
         self.inner.is_idle()
     }
 }
-impl<E, T, F> ExactBytesEncode for MapFrom<E, T, F>
+impl<E, T, F> SizedEncode for MapFrom<E, T, F>
 where
-    E: ExactBytesEncode,
+    E: SizedEncode,
     F: Fn(T) -> E::Item,
 {
     fn exact_requiring_bytes(&self) -> u64 {
@@ -256,6 +318,21 @@ pub struct TryMapFrom<C, T, E, F> {
     _phantom: PhantomData<(T, E)>,
 }
 impl<C, T, E, F> TryMapFrom<C, T, E, F> {
+    /// Returns a reference to the inner encoder.
+    pub fn inner_ref(&self) -> &C {
+        &self.inner
+    }
+
+    /// Returns a mutable reference to the inner encoder.
+    pub fn inner_mut(&mut self) -> &mut C {
+        &mut self.inner
+    }
+
+    /// Takes ownership of this instance and returns the inner encoder.
+    pub fn into_inner(self) -> C {
+        self.inner
+    }
+
     pub(crate) fn new(inner: C, try_from: F) -> Self {
         TryMapFrom {
             inner,
@@ -289,9 +366,9 @@ where
         self.inner.is_idle()
     }
 }
-impl<C, T, E, F> ExactBytesEncode for TryMapFrom<C, T, E, F>
+impl<C, T, E, F> SizedEncode for TryMapFrom<C, T, E, F>
 where
-    C: ExactBytesEncode,
+    C: SizedEncode,
     F: Fn(T) -> std::result::Result<C::Item, E>,
     Error: From<E>,
 {
@@ -421,14 +498,19 @@ impl<D> Omittable<D> {
 impl<D: Decode> Decode for Omittable<D> {
     type Item = Option<D::Item>;
 
-    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<(usize, Option<Self::Item>)> {
+    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<usize> {
         if self.do_omit {
-            Ok((0, Some(None)))
+            Ok(0)
         } else {
-            match track!(self.inner.decode(buf, eos))? {
-                (size, Some(item)) => Ok((size, Some(Some(item)))),
-                (size, None) => Ok((size, None)),
-            }
+            track!(self.inner.decode(buf, eos))
+        }
+    }
+
+    fn finish_decoding(&mut self) -> Result<Self::Item> {
+        if self.do_omit {
+            Ok(None)
+        } else {
+            track!(self.inner.finish_decoding()).map(Some)
         }
     }
 
@@ -439,6 +521,10 @@ impl<D: Decode> Decode for Omittable<D> {
             self.inner.requiring_bytes()
         }
     }
+
+    fn is_idle(&self) -> bool {
+        self.do_omit || self.inner.is_idle()
+    }
 }
 
 /// Combinator for representing an optional encoder.
@@ -486,7 +572,7 @@ impl<E: Encode> Encode for Optional<E> {
         self.0.is_idle()
     }
 }
-impl<E: ExactBytesEncode> ExactBytesEncode for Optional<E> {
+impl<E: SizedEncode> SizedEncode for Optional<E> {
     fn exact_requiring_bytes(&self) -> u64 {
         self.0.exact_requiring_bytes()
     }
@@ -502,7 +588,7 @@ impl<E: ExactBytesEncode> ExactBytesEncode for Optional<E> {
 pub struct Collect<D, T> {
     inner: D,
     items: T,
-    is_decoding: bool,
+    eos: bool,
 }
 impl<D, T: Default> Collect<D, T> {
     /// Returns a reference to the inner decoder.
@@ -524,7 +610,7 @@ impl<D, T: Default> Collect<D, T> {
         Collect {
             inner,
             items: T::default(),
-            is_decoding: false,
+            eos: false,
         }
     }
 }
@@ -535,33 +621,41 @@ where
 {
     type Item = T;
 
-    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<(usize, Option<Self::Item>)> {
+    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<usize> {
+        if self.eos {
+            return Ok(0);
+        }
+
         let mut offset = 0;
         while offset < buf.len() {
-            let (size, item) = track!(self.inner.decode(&buf[offset..], eos))?;
-            offset += size;
-            if size != 0 {
-                self.is_decoding = true;
-            }
+            bytecodec_try_decode!(self.inner, offset, buf, eos);
 
-            if item.is_some() {
-                self.items.extend(item);
-                self.is_decoding = false;
-            } else {
-                break;
-            }
+            let item = track!(self.inner.finish_decoding())?;
+            self.items.extend(iter::once(item));
         }
-        if offset == buf.len() && eos.is_reached() {
-            track_assert!(!self.is_decoding, ErrorKind::UnexpectedEos);
-            let items = mem::replace(&mut self.items, T::default());
-            Ok((offset, Some(items)))
-        } else {
-            Ok((offset, None))
+        if eos.is_reached() {
+            self.eos = true;
         }
+        Ok(offset)
+    }
+
+    fn finish_decoding(&mut self) -> Result<Self::Item> {
+        track_assert!(self.eos, ErrorKind::IncompleteDecoding);
+        self.eos = false;
+        let items = mem::replace(&mut self.items, T::default());
+        Ok(items)
     }
 
     fn requiring_bytes(&self) -> ByteCount {
-        self.inner.requiring_bytes()
+        if self.eos {
+            ByteCount::Finite(0)
+        } else {
+            self.inner.requiring_bytes()
+        }
+    }
+
+    fn is_idle(&self) -> bool {
+        self.eos
     }
 }
 
@@ -592,12 +686,12 @@ impl<C> Length<C> {
     ///
     /// # Errors
     ///
-    /// If it is in the middle of decoding an item, it willl return an `ErrorKind::Other` error.
+    /// If it is in the middle of decoding an item, it willl return an `ErrorKind::IncompleteDecoding` error.
     pub fn set_expected_bytes(&mut self, bytes: u64) -> Result<()> {
         track_assert_eq!(
             self.remaining_bytes,
             self.expected_bytes,
-            ErrorKind::Other,
+            ErrorKind::IncompleteDecoding,
             "An item is being decoded"
         );
         self.expected_bytes = bytes;
@@ -628,32 +722,33 @@ impl<C> Length<C> {
 impl<D: Decode> Decode for Length<D> {
     type Item = D::Item;
 
-    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<(usize, Option<Self::Item>)> {
+    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<usize> {
         let limit = cmp::min(buf.len() as u64, self.remaining_bytes) as usize;
         let required = self.remaining_bytes - limit as u64;
         let expected_eos = Eos::with_remaining_bytes(ByteCount::Finite(required));
         if let Some(remaining) = eos.remaining_bytes().to_u64() {
             track_assert!(remaining >= required, ErrorKind::UnexpectedEos; remaining, required);
         }
-        let (size, item) = track!(self.inner.decode(&buf[..limit], expected_eos))?;
+
+        let size = track!(self.inner.decode(&buf[..limit], expected_eos))?;
         self.remaining_bytes -= size as u64;
-        if item.is_some() {
-            track_assert_eq!(
-                self.remaining_bytes,
-                0,
-                ErrorKind::Other,
-                "Decoder consumes too few bytes"
-            );
-            self.remaining_bytes = self.expected_bytes
-        } else {
-            track_assert_ne!(self.remaining_bytes, 0, ErrorKind::Other);
-        }
-        Ok((size, item))
+        Ok(size)
+    }
+
+    fn finish_decoding(&mut self) -> Result<Self::Item> {
+        track_assert_eq!(self.remaining_bytes, 0, ErrorKind::IncompleteDecoding);
+        self.remaining_bytes = self.expected_bytes;
+
+        track!(self.inner.finish_decoding())
     }
 
     fn requiring_bytes(&self) -> ByteCount {
         ByteCount::Finite(self.remaining_bytes)
     }
+
+    fn is_idle(&self) -> bool {
+        self.remaining_bytes == 0
+    }
 }
 impl<E: Encode> Encode for Length<E> {
     type Item = E::Item;
@@ -699,7 +794,7 @@ impl<E: Encode> Encode for Length<E> {
         self.remaining_bytes == 0
     }
 }
-impl<E: Encode> ExactBytesEncode for Length<E> {
+impl<E: Encode> SizedEncode for Length<E> {
     fn exact_requiring_bytes(&self) -> u64 {
         self.remaining_bytes
     }
@@ -755,29 +850,25 @@ where
 {
     type Item = T;
 
-    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<(usize, Option<Self::Item>)> {
-        if self.remaining_items == 0 {
-            return Ok((0, Some(T::default())));
-        }
-
+    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<usize> {
         let mut offset = 0;
-        while offset < buf.len() {
-            let (size, item) = track!(self.inner.decode(&buf[offset..], eos))?;
-            offset += size;
+        while self.remaining_items != 0 && offset < buf.len() {
+            bytecodec_try_decode!(self.inner, offset, buf, eos);
 
-            if item.is_some() {
-                self.items.extend(item);
-                self.remaining_items -= 1;
-                if self.remaining_items == 0 {
-                    let items = mem::replace(&mut self.items, T::default());
-                    return Ok((offset, Some(items)));
-                }
-            } else {
-                break;
-            }
+            let item = track!(self.inner.finish_decoding())?;
+            self.items.extend(iter::once(item));
+            self.remaining_items -= 1;
         }
-        track_assert!(!eos.is_reached(), ErrorKind::UnexpectedEos);
-        Ok((offset, None))
+        if self.remaining_items != 0 {
+            track_assert!(!eos.is_reached(), ErrorKind::UnexpectedEos);
+        }
+        Ok(offset)
+    }
+
+    fn finish_decoding(&mut self) -> Result<Self::Item> {
+        track_assert_eq!(self.remaining_items, 0, ErrorKind::IncompleteDecoding);
+        let items = mem::replace(&mut self.items, T::default());
+        Ok(items)
     }
 
     fn requiring_bytes(&self) -> ByteCount {
@@ -787,61 +878,22 @@ where
             self.inner.requiring_bytes()
         }
     }
+
+    fn is_idle(&self) -> bool {
+        self.remaining_items == 0
+    }
 }
 
 /// Combinator which tries to convert decoded values by calling the specified function.
 ///
 /// This is created by calling `DecodeExt::try_map` method.
 #[derive(Debug)]
-pub struct TryMap<D, F, T, E> {
+pub struct TryMap<D, T, E, F> {
     inner: D,
     try_map: F,
     _phantom: PhantomData<(T, E)>,
 }
-impl<D, F, T, E> TryMap<D, F, T, E> {
-    pub(crate) fn new(inner: D, try_map: F) -> Self {
-        TryMap {
-            inner,
-            try_map,
-            _phantom: PhantomData,
-        }
-    }
-}
-impl<D, F, T, E> Decode for TryMap<D, F, T, E>
-where
-    D: Decode,
-    F: Fn(D::Item) -> std::result::Result<T, E>,
-    Error: From<E>,
-{
-    type Item = T;
-
-    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<(usize, Option<Self::Item>)> {
-        match track!(self.inner.decode(buf, eos))? {
-            (size, Some(item)) => {
-                let item = track!((self.try_map)(item).map_err(Error::from))?;
-                Ok((size, Some(item)))
-            }
-            (size, None) => Ok((size, None)),
-        }
-    }
-
-    fn requiring_bytes(&self) -> ByteCount {
-        self.inner.requiring_bytes()
-    }
-}
-
-/// Combinator for skipping the remaining bytes in an input byte sequence
-/// after decoding an item by using `D`.
-#[derive(Debug, Default)]
-pub struct SkipRemaining<D: Decode> {
-    inner: D,
-    item: Option<D::Item>,
-}
-impl<D: Decode> SkipRemaining<D> {
-    pub(crate) fn new(inner: D) -> Self {
-        SkipRemaining { inner, item: None }
-    }
-
+impl<D, T, E, F> TryMap<D, T, E, F> {
     /// Returns a reference to the inner decoder.
     pub fn inner_ref(&self) -> &D {
         &self.inner
@@ -856,34 +908,38 @@ impl<D: Decode> SkipRemaining<D> {
     pub fn into_inner(self) -> D {
         self.inner
     }
-}
-impl<D: Decode> Decode for SkipRemaining<D> {
-    type Item = D::Item;
 
-    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<(usize, Option<Self::Item>)> {
-        track_assert!(
-            !eos.remaining_bytes().is_infinite(),
-            ErrorKind::InvalidInput,
-            "Cannot skip infinity byte stream"
-        );
-
-        if self.item.is_none() {
-            let (size, item) = track!(self.inner.decode(buf, eos))?;
-            self.item = item;
-            Ok((size, None))
-        } else if eos.is_reached() {
-            Ok((buf.len(), self.item.take()))
-        } else {
-            Ok((buf.len(), None))
+    pub(crate) fn new(inner: D, try_map: F) -> Self {
+        TryMap {
+            inner,
+            try_map,
+            _phantom: PhantomData,
         }
     }
+}
+impl<D, T, E, F> Decode for TryMap<D, T, E, F>
+where
+    D: Decode,
+    F: Fn(D::Item) -> std::result::Result<T, E>,
+    Error: From<E>,
+{
+    type Item = T;
+
+    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<usize> {
+        track!(self.inner.decode(buf, eos))
+    }
+
+    fn finish_decoding(&mut self) -> Result<Self::Item> {
+        let item = track!(self.inner.finish_decoding())?;
+        track!((self.try_map)(item).map_err(Error::from))
+    }
 
     fn requiring_bytes(&self) -> ByteCount {
-        if self.item.is_none() {
-            self.inner.requiring_bytes()
-        } else {
-            ByteCount::Infinite
-        }
+        self.inner.requiring_bytes()
+    }
+
+    fn is_idle(&self) -> bool {
+        self.inner.is_idle()
     }
 }
 
@@ -951,7 +1007,7 @@ impl<C> MaxBytes<C> {
 impl<D: Decode> Decode for MaxBytes<D> {
     type Item = D::Item;
 
-    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<(usize, Option<Self::Item>)> {
+    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<usize> {
         match eos.remaining_bytes() {
             ByteCount::Infinite => {
                 track_panic!(ErrorKind::InvalidInput, "Max bytes limit exceeded";
@@ -971,17 +1027,22 @@ impl<D: Decode> Decode for MaxBytes<D> {
             }
         }
 
-        let (size, item) = track!(self.inner.decode(buf, eos))?;
+        let size = track!(self.inner.decode(buf, eos))?;
         self.consumed_bytes += size as u64;
-        if item.is_some() {
-            self.consumed_bytes = 0;
-        }
-        Ok((size, item))
+        Ok(size)
+    }
+
+    fn finish_decoding(&mut self) -> Result<Self::Item> {
+        track!(self.inner.finish_decoding())
     }
 
     fn requiring_bytes(&self) -> ByteCount {
         self.inner.requiring_bytes()
     }
+
+    fn is_idle(&self) -> bool {
+        self.inner.is_idle()
+    }
 }
 impl<E: Encode> Encode for MaxBytes<E> {
     type Item = E::Item;
@@ -1010,55 +1071,21 @@ impl<E: Encode> Encode for MaxBytes<E> {
         self.inner.is_idle()
     }
 }
-impl<E: ExactBytesEncode> ExactBytesEncode for MaxBytes<E> {
+impl<E: SizedEncode> SizedEncode for MaxBytes<E> {
     fn exact_requiring_bytes(&self) -> u64 {
         self.inner.exact_requiring_bytes()
     }
 }
 
-/// Combinator for declaring an assertion about decoded items.
+/// Combinator for pre-encoding items when `start_encoding` method is called.
 ///
-/// This created by calling `DecodeExt::assert` method.
-#[derive(Debug)]
-pub struct Assert<D, F> {
-    inner: D,
-    assert: F,
-}
-impl<D, F> Assert<D, F> {
-    pub(crate) fn new(inner: D, assert: F) -> Self {
-        Assert { inner, assert }
-    }
-}
-impl<D: Decode, F> Decode for Assert<D, F>
-where
-    F: for<'a> Fn(&'a D::Item) -> bool,
-{
-    type Item = D::Item;
-
-    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<(usize, Option<Self::Item>)> {
-        let (size, item) = track!(self.inner.decode(buf, eos))?;
-        if let Some(ref item) = item {
-            track_assert!((self.assert)(item), ErrorKind::InvalidInput);
-        }
-        Ok((size, item))
-    }
-
-    fn requiring_bytes(&self) -> ByteCount {
-        self.inner.requiring_bytes()
-    }
-}
-
-/// Combinator that keeps writing padding byte until it reaches EOS
-/// after encoding of `E`'s item has been completed.
-///
-/// This is created by calling `EncodeExt::padding` method.
+/// This is created by calling `EncodeExt::pre_encode` method.
 #[derive(Debug, Default)]
-pub struct Padding<E> {
+pub struct PreEncode<E> {
     inner: E,
-    padding_byte: u8,
-    eos_reached: bool,
+    pre_encoded: BytesEncoder<Vec<u8>>,
 }
-impl<E> Padding<E> {
+impl<E> PreEncode<E> {
     /// Returns a reference to the inner encoder.
     pub fn inner_ref(&self) -> &E {
         &self.inner
@@ -1074,118 +1101,6 @@ impl<E> Padding<E> {
         self.inner
     }
 
-    pub(crate) fn new(inner: E, padding_byte: u8) -> Self {
-        Padding {
-            inner,
-            padding_byte,
-            eos_reached: true,
-        }
-    }
-}
-impl<E: Encode> Encode for Padding<E> {
-    type Item = E::Item;
-
-    fn encode(&mut self, buf: &mut [u8], eos: Eos) -> Result<usize> {
-        let mut offset = 0;
-        bytecodec_try_encode!(self.inner, offset, buf, eos);
-
-        for b in (&mut buf[offset..]).iter_mut() {
-            *b = self.padding_byte;
-        }
-        self.eos_reached = eos.is_reached();
-        Ok(buf.len())
-    }
-
-    fn start_encoding(&mut self, item: Self::Item) -> Result<()> {
-        track_assert!(self.is_idle(), ErrorKind::EncoderFull);
-        self.eos_reached = false;
-        track!(self.inner.start_encoding(item))
-    }
-
-    fn requiring_bytes(&self) -> ByteCount {
-        ByteCount::Infinite
-    }
-
-    fn is_idle(&self) -> bool {
-        self.eos_reached
-    }
-}
-
-/// Combinator for adding prefix items.
-///
-/// This is created by calling `EncodeExt::with_prefix` method.
-#[derive(Debug)]
-pub struct WithPrefix<E0, E1, F> {
-    body_encoder: E0,
-    prefix_encoder: E1,
-    with_prefix: F,
-}
-impl<E0, E1, F> WithPrefix<E0, E1, F> {
-    pub(crate) fn new(body_encoder: E0, prefix_encoder: E1, with_prefix: F) -> Self {
-        WithPrefix {
-            body_encoder,
-            prefix_encoder,
-            with_prefix,
-        }
-    }
-}
-impl<E0, E1, F> Encode for WithPrefix<E0, E1, F>
-where
-    E0: Encode,
-    E1: Encode,
-    F: Fn(&E0) -> E1::Item,
-{
-    type Item = E0::Item;
-
-    fn encode(&mut self, buf: &mut [u8], eos: Eos) -> Result<usize> {
-        let mut offset = 0;
-        bytecodec_try_encode!(self.prefix_encoder, offset, buf, eos);
-        bytecodec_try_encode!(self.body_encoder, offset, buf, eos);
-        Ok(offset)
-    }
-
-    fn start_encoding(&mut self, item: Self::Item) -> Result<()> {
-        track_assert!(self.is_idle(), ErrorKind::EncoderFull);
-        track!(self.body_encoder.start_encoding(item))?;
-        let prefix_item = (self.with_prefix)(&self.body_encoder);
-        track!(self.prefix_encoder.start_encoding(prefix_item))?;
-        Ok(())
-    }
-
-    fn requiring_bytes(&self) -> ByteCount {
-        let x = self.prefix_encoder.requiring_bytes();
-        let y = self.body_encoder.requiring_bytes();
-        match (x, y) {
-            (ByteCount::Finite(x), ByteCount::Finite(y)) => ByteCount::Finite(x + y),
-            (ByteCount::Infinite, _) | (_, ByteCount::Infinite) => ByteCount::Infinite,
-            (ByteCount::Unknown, _) | (_, ByteCount::Unknown) => ByteCount::Unknown,
-        }
-    }
-
-    fn is_idle(&self) -> bool {
-        self.prefix_encoder.is_idle() && self.body_encoder.is_idle()
-    }
-}
-impl<E0, E1, F> ExactBytesEncode for WithPrefix<E0, E1, F>
-where
-    E0: ExactBytesEncode,
-    E1: ExactBytesEncode,
-    F: Fn(&E0) -> E1::Item,
-{
-    fn exact_requiring_bytes(&self) -> u64 {
-        self.prefix_encoder.exact_requiring_bytes() + self.body_encoder.exact_requiring_bytes()
-    }
-}
-
-/// Combinator for pre-encoding items when `start_encoding` method is called.
-///
-/// This is created by calling `EncodeExt::pre_encode` method.
-#[derive(Debug, Default)]
-pub struct PreEncode<E> {
-    inner: E,
-    pre_encoded: BytesEncoder<Vec<u8>>,
-}
-impl<E> PreEncode<E> {
     pub(crate) fn new(inner: E) -> Self {
         PreEncode {
             inner,
@@ -1207,14 +1122,14 @@ impl<E: Encode> Encode for PreEncode<E> {
     }
 
     fn requiring_bytes(&self) -> ByteCount {
-        ByteCount::Finite(self.exact_requiring_bytes())
+        self.pre_encoded.requiring_bytes()
     }
 
     fn is_idle(&self) -> bool {
         self.pre_encoded.is_idle()
     }
 }
-impl<E: Encode> ExactBytesEncode for PreEncode<E> {
+impl<E: Encode> SizedEncode for PreEncode<E> {
     fn exact_requiring_bytes(&self) -> u64 {
         self.pre_encoded.exact_requiring_bytes()
     }
@@ -1274,17 +1189,25 @@ impl<T> Slice<T> {
 impl<D: Decode> Decode for Slice<D> {
     type Item = D::Item;
 
-    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<(usize, Option<Self::Item>)> {
+    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<usize> {
         let limit = cmp::min(buf.len() as u64, self.consumable_bytes) as usize;
         let eos = eos.back((buf.len() - limit) as u64);
-        let (size, item) = track!(self.inner.decode(&buf[..limit], eos))?;
+        let size = track!(self.inner.decode(&buf[..limit], eos))?;
         self.consumable_bytes -= size as u64;
-        Ok((size, item))
+        Ok(size)
+    }
+
+    fn finish_decoding(&mut self) -> Result<Self::Item> {
+        track!(self.inner.finish_decoding())
     }
 
     fn requiring_bytes(&self) -> ByteCount {
         self.inner.requiring_bytes()
     }
+
+    fn is_idle(&self) -> bool {
+        self.inner.is_idle()
+    }
 }
 impl<E: Encode> Encode for Slice<E> {
     type Item = E::Item;
@@ -1309,59 +1232,7 @@ impl<E: Encode> Encode for Slice<E> {
         self.inner.requiring_bytes()
     }
 }
-impl<E: ExactBytesEncode> ExactBytesEncode for Slice<E> {
-    fn exact_requiring_bytes(&self) -> u64 {
-        self.inner.exact_requiring_bytes()
-    }
-}
-
-/// Combinator for representing encoders that cannot accept any more items.
-///
-/// This is created by calling `EncodeExt::last`.
-#[derive(Debug, Default)]
-pub struct Last<E> {
-    inner: E,
-}
-impl<E> Last<E> {
-    /// Returns a reference to the inner encoder.
-    pub fn inner_ref(&self) -> &E {
-        &self.inner
-    }
-
-    /// Returns a mutable reference to the inner encoder.
-    pub fn inner_mut(&mut self) -> &mut E {
-        &mut self.inner
-    }
-
-    /// Takes ownership of this instance and returns the inner encoder.
-    pub fn into_inner(self) -> E {
-        self.inner
-    }
-
-    pub(crate) fn new(inner: E) -> Self {
-        Last { inner }
-    }
-}
-impl<E: Encode> Encode for Last<E> {
-    type Item = Never;
-
-    fn encode(&mut self, buf: &mut [u8], eos: Eos) -> Result<usize> {
-        track!(self.inner.encode(buf, eos))
-    }
-
-    fn start_encoding(&mut self, _item: Self::Item) -> Result<()> {
-        unreachable!()
-    }
-
-    fn is_idle(&self) -> bool {
-        self.inner.is_idle()
-    }
-
-    fn requiring_bytes(&self) -> ByteCount {
-        self.inner.requiring_bytes()
-    }
-}
-impl<E: ExactBytesEncode> ExactBytesEncode for Last<E> {
+impl<E: SizedEncode> SizedEncode for Slice<E> {
     fn exact_requiring_bytes(&self) -> u64 {
         self.inner.exact_requiring_bytes()
     }
@@ -1369,13 +1240,13 @@ impl<E: ExactBytesEncode> ExactBytesEncode for Last<E> {
 
 /// Combinator for representing encoders that accepts only one additional item.
 ///
-/// This is created by calling `EncodeExt::last_item`.
+/// This is created by calling `EncodeExt::last`.
 #[derive(Debug, Default)]
-pub struct LastItem<E: Encode> {
+pub struct Last<E: Encode> {
     inner: E,
     item: Option<E::Item>,
 }
-impl<E: Encode> LastItem<E> {
+impl<E: Encode> Last<E> {
     /// Returns a reference to the inner encoder.
     pub fn inner_ref(&self) -> &E {
         &self.inner
@@ -1392,13 +1263,13 @@ impl<E: Encode> LastItem<E> {
     }
 
     pub(crate) fn new(inner: E, item: E::Item) -> Self {
-        LastItem {
+        Last {
             inner,
             item: Some(item),
         }
     }
 }
-impl<E: Encode> Encode for LastItem<E> {
+impl<E: Encode> Encode for Last<E> {
     type Item = Never;
 
     fn encode(&mut self, buf: &mut [u8], eos: Eos) -> Result<usize> {
@@ -1426,6 +1297,99 @@ impl<E: Encode> Encode for LastItem<E> {
         }
     }
 }
+impl<E: SizedEncode> SizedEncode for Last<E> {
+    fn exact_requiring_bytes(&self) -> u64 {
+        self.inner.exact_requiring_bytes()
+    }
+}
+
+/// Combinator that enables to peek decoded items before calling `finish_decoding` method.
+///
+/// This is created by calling `DecodeExt::peekable` method.
+pub struct Peekable<D: Decode> {
+    inner: D,
+    item: Option<D::Item>,
+}
+impl<D: Decode> Peekable<D> {
+    /// Returns a reference to the last item decoded (but not `finish_decoding` called yet) by the decoder.
+    pub fn peek(&self) -> Option<&D::Item> {
+        self.item.as_ref()
+    }
+
+    /// Returns a mutable reference to the last item decoded (but not `finish_decoding` called yet) by the decoder.
+    pub fn peek_mut(&mut self) -> Option<&mut D::Item> {
+        self.item.as_mut()
+    }
+
+    /// Returns a reference to the inner decoder.
+    pub fn inner_ref(&self) -> &D {
+        &self.inner
+    }
+
+    /// Returns a mutable reference to the inner decoder.
+    pub fn inner_mut(&mut self) -> &mut D {
+        &mut self.inner
+    }
+
+    /// Takes ownership of this instance and returns the inner decoder.
+    pub fn into_inner(self) -> D {
+        self.inner
+    }
+
+    pub(crate) fn new(inner: D) -> Self {
+        Peekable { inner, item: None }
+    }
+}
+impl<D: Decode + fmt::Debug> fmt::Debug for Peekable<D> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        write!(
+            f,
+            "Peekable {{ inner: {:?}, item.is_some(): {:?} }}",
+            self.inner,
+            self.item.is_some()
+        )
+    }
+}
+impl<D: Decode + Default> Default for Peekable<D> {
+    fn default() -> Self {
+        Peekable {
+            inner: D::default(),
+            item: None,
+        }
+    }
+}
+impl<D: Decode> Decode for Peekable<D> {
+    type Item = D::Item;
+
+    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<usize> {
+        if self.item.is_none() {
+            let size = track!(self.inner.decode(buf, eos))?;
+            if self.inner.is_idle() {
+                self.item = Some(track!(self.inner.finish_decoding())?);
+            }
+            Ok(size)
+        } else {
+            Ok(0)
+        }
+    }
+
+    fn finish_decoding(&mut self) -> Result<Self::Item> {
+        let item = track_assert_some!(self.item.take(), ErrorKind::IncompleteDecoding);
+        Ok(item)
+    }
+
+    fn requiring_bytes(&self) -> ByteCount {
+        if self.item.is_some() {
+            ByteCount::Finite(0)
+        } else {
+            self.inner.requiring_bytes()
+        }
+    }
+
+    fn is_idle(&self) -> bool {
+        self.item.is_some()
+    }
+}
 
 /// Combinator for ignoring EOS if there is no item being decoded.
 ///
@@ -1461,26 +1425,30 @@ impl<D> MaybeEos<D> {
 impl<D: Decode> Decode for MaybeEos<D> {
     type Item = D::Item;
 
-    fn decode(&mut self, buf: &[u8], mut eos: Eos) -> Result<(usize, Option<Self::Item>)> {
+    fn decode(&mut self, buf: &[u8], mut eos: Eos) -> Result<usize> {
         if !self.started && buf.is_empty() && eos.is_reached() {
             eos = Eos::new(false);
         }
 
-        let (size, item) = track!(self.inner.decode(buf, eos))?;
-        if let Some(item) = item {
-            self.started = false;
-            Ok((size, Some(item)))
-        } else {
-            if size != 0 {
-                self.started = true;
-            }
-            Ok((size, None))
+        let size = track!(self.inner.decode(buf, eos))?;
+        if size != 0 {
+            self.started = true;
         }
+        Ok(size)
+    }
+
+    fn finish_decoding(&mut self) -> Result<Self::Item> {
+        self.started = false;
+        track!(self.inner.finish_decoding())
     }
 
     fn requiring_bytes(&self) -> ByteCount {
         self.inner.requiring_bytes()
     }
+
+    fn is_idle(&self) -> bool {
+        self.inner.is_idle()
+    }
 }
 
 #[cfg(test)]
@@ -1488,6 +1456,7 @@ mod test {
     use bytes::{Utf8Decoder, Utf8Encoder};
     use fixnum::{U16beDecoder, U8Decoder, U8Encoder};
     use io::{IoDecodeExt, IoEncodeExt};
+    use tuple::TupleDecoder;
     use {Decode, DecodeExt, Encode, EncodeExt, Eos, ErrorKind};
 
     #[test]
@@ -1549,15 +1518,6 @@ mod test {
         assert_eq!(*error.kind(), ErrorKind::InvalidInput);
     }
 
-    #[test]
-    fn padding_works() {
-        let mut output = Vec::new();
-        let mut encoder = U8Encoder::new().padding(9).length(3);
-        encoder.start_encoding(3).unwrap();
-        track_try_unwrap!(encoder.encode_all(&mut output));
-        assert_eq!(output, [3, 9, 9]);
-    }
-
     #[test]
     fn repeat_works() {
         let mut output = Vec::new();
@@ -1591,23 +1551,17 @@ mod test {
         let input = b"fboaor";
         let mut offset = 0;
 
-        let mut last_item0 = None;
-        let mut last_item1 = None;
         for _ in 0..3 {
             decoder0.set_consumable_bytes(1);
-            let (size, item) = track_try_unwrap!(decoder0.decode(&input[offset..], eos));
-            offset += size;
-            last_item0 = item;
+            offset += track_try_unwrap!(decoder0.decode(&input[offset..], eos));
 
             decoder1.set_consumable_bytes(1);
-            let (size, item) = track_try_unwrap!(decoder1.decode(&input[offset..], eos));
-            offset += size;
-            last_item1 = item;
+            offset += track_try_unwrap!(decoder1.decode(&input[offset..], eos));
         }
 
         assert_eq!(offset, input.len());
-        assert_eq!(last_item0, Some("foo".to_owned()));
-        assert_eq!(last_item1, Some("bar".to_owned()));
+        assert_eq!(decoder0.finish_decoding().ok(), Some("foo".to_owned()));
+        assert_eq!(decoder1.finish_decoding().ok(), Some("bar".to_owned()));
     }
 
     #[test]
@@ -1643,8 +1597,8 @@ mod test {
     fn and_then_works() {
         let mut decoder =
             U8Decoder::new().and_then(|len| Utf8Decoder::new().length(u64::from(len)));
-        let (_, item) = track_try_unwrap!(decoder.decode(b"\x03foo", Eos::new(false)));
-        assert_eq!(item, Some("foo".to_owned()));
+        track_try_unwrap!(decoder.decode(b"\x03foo", Eos::new(false)));
+        assert_eq!(track_try_unwrap!(decoder.finish_decoding()), "foo");
     }
 
     #[test]
@@ -1659,4 +1613,15 @@ mod test {
         assert!(decoder.decode(&[1][..], Eos::new(false)).is_ok());
         assert!(decoder.decode(&[][..], Eos::new(true)).is_err());
     }
+
+    #[test]
+    fn peekable_works() {
+        let mut decoder =
+            TupleDecoder::new((U8Decoder::new(), U8Decoder::new(), U8Decoder::new())).peekable();
+        let size = decoder.decode(b"foo", Eos::new(false)).unwrap();
+        assert_eq!(size, 3);
+        assert_eq!(decoder.peek(), Some(&(b'f', b'o', b'o')));
+        assert_eq!(decoder.finish_decoding().unwrap(), (b'f', b'o', b'o'));
+        assert_eq!(decoder.peek(), None);
+    }
 }
