@@ -6,19 +6,19 @@ use {ByteCount, Decode, Encode, Eos, Error, ErrorKind, Result};
 
 /// An extension of `Decode` trait to aid decodings involving I/O.
 pub trait IoDecodeExt: Decode {
-    /// Decodes an item from the given read buffer.
-    fn decode_from_read_buf<B>(&mut self, buf: &mut ReadBuf<B>) -> Result<Option<Self::Item>>
+    /// Consumes bytes from the given read buffer and proceeds the decoding process.
+    fn decode_from_read_buf<B>(&mut self, buf: &mut ReadBuf<B>) -> Result<()>
     where
         B: AsRef<[u8]>,
     {
         let eos = Eos::new(buf.stream_state.is_eos());
-        let (size, item) = track!(self.decode(&buf.inner.as_ref()[buf.head..buf.tail], eos))?;
+        let size = track!(self.decode(&buf.inner.as_ref()[buf.head..buf.tail], eos))?;
         buf.head += size;
         if buf.head == buf.tail {
             buf.head = 0;
             buf.tail = 0;
         }
-        Ok(item)
+        Ok(())
     }
 
     /// Decodes an item from the given reader.
@@ -41,9 +41,10 @@ pub trait IoDecodeExt: Decode {
                 Eos::new(false)
             };
 
-            let (consumed, item) = track!(self.decode(&buf[..size], eos))?;
-            track_assert_eq!(consumed, size, ErrorKind::Other; item.is_some(), eos);
-            if let Some(item) = item {
+            let consumed = track!(self.decode(&buf[..size], eos))?;
+            track_assert_eq!(consumed, size, ErrorKind::InconsistentState; self.is_idle(), eos);
+            if self.is_idle() {
+                let item = track!(self.finish_decoding())?;
                 return Ok(item);
             }
         }
@@ -439,8 +440,8 @@ mod test {
         assert_eq!(buf.stream_state(), StreamState::Eos);
 
         let mut decoder = Utf8Decoder::new();
-        let item = track_try_unwrap!(decoder.decode_from_read_buf(&mut buf));
-        assert_eq!(item, Some("foo".to_owned()));
+        track_try_unwrap!(decoder.decode_from_read_buf(&mut buf));
+        assert_eq!(track_try_unwrap!(decoder.finish_decoding()), "foo");
     }
 
     #[test]
